<%inherit file="base.htm" />

<%block name="content">
    <div class="content">
        <h3>Congratulations -- you've successfully created a new django-mako-plus app!</h3>
        <h4>Next Up: Go through the django-mako-plus tutorial and add Javascript, CSS, and urlparams to this page.</h4>
        <h4 class="utc-time">Current time in UTC: ${ utc_time }</h4>
    </div>
</%block>